/* Autogenerated by build_info.pl - DO NOT EDIT */

/* This stores global variables that are initialized with
   preprocessor declarations for output with the --version flag.

   Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003,
   2004, 2005, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.  */

#include "wget.h"
#include <stdio.h>
#include "version.h"

const char *compiled_features[] =
{

#if defined HAVE_LIBCARES
  "+cares",
#else
  "-cares",
#endif

#if defined ENABLE_DIGEST
  "+digest",
#else
  "-digest",
#endif

#if defined HAVE_GPGME
  "+gpgme",
#else
  "-gpgme",
#endif

#if defined HAVE_SSL
  "+https",
#else
  "-https",
#endif

#if defined ENABLE_IPV6
  "+ipv6",
#else
  "-ipv6",
#endif

#if defined ENABLE_IRI
  "+iri",
#else
  "-iri",
#endif

#if SIZEOF_OFF_T >= 8 || defined WINDOWS
  "+large-file",
#else
  "-large-file",
#endif

#if defined HAVE_METALINK
  "+metalink",
#else
  "-metalink",
#endif

#if defined ENABLE_NLS
  "+nls",
#else
  "-nls",
#endif

#if defined ENABLE_NTLM
  "+ntlm",
#else
  "-ntlm",
#endif

#if defined ENABLE_OPIE
  "+opie",
#else
  "-opie",
#endif

#if defined HAVE_LIBPSL
  "+psl",
#else
  "-psl",
#endif

#if defined HAVE_LIBSSL || defined HAVE_LIBSSL32
  "+ssl/openssl",
#elif defined HAVE_LIBGNUTLS
  "+ssl/gnutls",
#else
  "-ssl",
#endif


  /* sentinel value */
  NULL
};


