/* version.c */
/* Autogenerated by Makefile - DO NOT EDIT */

#include "version.h"
const char *version_string = "1.21.3";
const char *compilation_string = "aarch64-xilinx-linux-gcc -mcpu=cortex-a72.cortex-a53 -march=armv8-a+crc -fstack-protector-strong -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  -DHAVE_CONFIG_H -DSYSTEM_WGETRC=\"/etc/wgetrc\" -DLOCALEDIR=\"/usr/share/locale\" -I. -I../../wget-1.21.3/src -I../lib -I../../wget-1.21.3/lib -DNDEBUG -O2 -pipe -g -feliminate-unused-debug-types  ";
const char *link_string = "aarch64-xilinx-linux-gcc -mcpu=cortex-a72.cortex-a53 -march=armv8-a+crc -fstack-protector-strong -O2 -D_FORTIFY_SOURCE=2 -Wformat -Wformat-security -Werror=format-security  -DNDEBUG -O2 -pipe -g -feliminate-unused-debug-types  -Wl,-O1 -Wl,--hash-style=gnu -Wl,--as-needed  -Wl,-z,relro,-z,now -lpcre -lnettle -lgnutls -lz ../lib/libgnu.a -lunistring ";
